package com.sun.corba.se.impl.transport;

import com.sun.corba.se.impl.orbutil.ORBUtility;
import com.sun.org.omg.SendingContext.CodeBase;
import sun.corba.OutputStreamFactory;
import com.sun.corba.se.impl.protocol.giopmsgheaders.Message;
import com.sun.corba.se.impl.protocol.giopmsgheaders.MessageBase;
import com.sun.corba.se.pept.transport.OutboundConnectionCache;
import com.sun.corba.se.pept.transport.InboundConnectionCache;
import org.omg.CORBA.CompletionStatus;
import org.omg.CORBA.SystemException;
import com.sun.corba.se.impl.encoding.OSFCodeSetRegistry;
import com.sun.corba.se.pept.transport.ResponseWaitingRoom;
import com.sun.corba.se.pept.transport.Connection;
import java.nio.channels.SelectableChannel;
import com.sun.corba.se.spi.orbutil.threadpool.NoSuchWorkQueueException;
import com.sun.corba.se.spi.orbutil.threadpool.NoSuchThreadPoolException;
import com.sun.corba.se.pept.encoding.InputObject;
import org.omg.CORBA.COMM_FAILURE;
import com.sun.corba.se.impl.encoding.CDROutputObject;
import com.sun.corba.se.pept.encoding.OutputObject;
import com.sun.corba.se.pept.transport.EventHandler;
import com.sun.corba.se.pept.transport.ContactInfo;
import java.io.InputStream;
import java.nio.ByteBuffer;
import com.sun.corba.se.pept.transport.Selector;
import com.sun.corba.se.spi.ior.iiop.GIOPVersion;
import org.omg.CORBA.INTERNAL;
import java.util.Collections;
import java.util.HashMap;
import java.io.IOException;
import java.net.InetSocketAddress;
import com.sun.corba.se.spi.orb.ORB;
import com.sun.corba.se.pept.protocol.MessageMediator;
import com.sun.corba.se.impl.encoding.CodeSetComponentInfo;
import com.sun.corba.se.spi.protocol.CorbaMessageMediator;
import com.sun.corba.se.spi.transport.ReadTimeouts;
import com.sun.corba.se.impl.logging.ORBUtilSystemException;
import com.sun.corba.se.impl.encoding.CachedCodeBase;
import com.sun.corba.se.spi.ior.IOR;
import java.util.Map;
import com.sun.corba.se.spi.transport.CorbaResponseWaitingRoom;
import java.net.Socket;
import com.sun.corba.se.pept.transport.ConnectionCache;
import com.sun.corba.se.pept.transport.Acceptor;
import com.sun.corba.se.spi.transport.CorbaContactInfo;
import java.nio.channels.SocketChannel;
import com.sun.corba.se.spi.orbutil.threadpool.Work;
import com.sun.corba.se.spi.transport.CorbaConnection;

public class SocketOrChannelConnectionImpl extends EventHandlerBase implements CorbaConnection, Work
{
    public static boolean dprintWriteLocks;
    protected long enqueueTime;
    protected SocketChannel socketChannel;
    protected CorbaContactInfo contactInfo;
    protected Acceptor acceptor;
    protected ConnectionCache connectionCache;
    protected Socket socket;
    protected long timeStamp;
    protected boolean isServer;
    protected int requestId;
    protected CorbaResponseWaitingRoom responseWaitingRoom;
    protected int state;
    protected Object stateEvent;
    protected Object writeEvent;
    protected boolean writeLocked;
    protected int serverRequestCount;
    Map serverRequestMap;
    protected boolean postInitialContexts;
    protected IOR codeBaseServerIOR;
    protected CachedCodeBase cachedCodeBase;
    protected ORBUtilSystemException wrapper;
    protected ReadTimeouts readTimeouts;
    protected boolean shouldReadGiopHeaderOnly;
    protected CorbaMessageMediator partialMessageMediator;
    protected CodeSetComponentInfo.CodeSetContext codeSetContext;
    protected MessageMediator clientReply_1_1;
    protected MessageMediator serverRequest_1_1;
    
    public SocketChannel getSocketChannel() {
        return this.socketChannel;
    }
    
    protected SocketOrChannelConnectionImpl(ORB orb) {
        this.timeStamp = 0L;
        this.isServer = false;
        this.requestId = 5;
        this.stateEvent = new Object();
        this.writeEvent = new Object();
        this.serverRequestCount = 0;
        this.serverRequestMap = null;
        this.postInitialContexts = false;
        this.cachedCodeBase = new CachedCodeBase(this);
        this.partialMessageMediator = null;
        this.codeSetContext = null;
        this.orb = orb;
        this.wrapper = ORBUtilSystemException.get(orb, "rpc.transport");
        this.setWork(this);
        this.responseWaitingRoom = new CorbaResponseWaitingRoomImpl(orb, this);
        this.setReadTimeouts(orb.getORBData().getTransportTCPReadTimeouts());
    }
    
    protected SocketOrChannelConnectionImpl(ORB orb, boolean useSelectThreadToWait, boolean useWorkerThreadForEvent) {
        this(orb);
        this.setUseSelectThreadToWait(useSelectThreadToWait);
        this.setUseWorkerThreadForEvent(useWorkerThreadForEvent);
    }
    
    public SocketOrChannelConnectionImpl(ORB orb, CorbaContactInfo contactInfo, boolean b, boolean b2, String s, String hostname, int n) {
        this(orb, b, b2);
        this.contactInfo = contactInfo;
        try {
            this.socket = orb.getORBData().getSocketFactory().createSocket(s, new InetSocketAddress(hostname, n));
            this.socketChannel = this.socket.getChannel();
            if (this.socketChannel != null) {
                this.socketChannel.configureBlocking(!b);
            }
            else {
                this.setUseSelectThreadToWait(false);
            }
            if (orb.transportDebugFlag) {
                this.dprint(".initialize: connection created: " + this.socket);
            }
        }
        catch (Throwable t) {
            throw this.wrapper.connectFailure(t, s, hostname, Integer.toString(n));
        }
        this.state = 1;
    }
    
    public SocketOrChannelConnectionImpl(ORB orb, CorbaContactInfo corbaContactInfo, String s, String s2, int n) {
        this(orb, corbaContactInfo, orb.getORBData().connectionSocketUseSelectThreadToWait(), orb.getORBData().connectionSocketUseWorkerThreadForEvent(), s, s2, n);
    }
    
    public SocketOrChannelConnectionImpl(ORB orb, Acceptor acceptor, Socket socket, boolean b, boolean b2) {
        this(orb, b, b2);
        this.socket = socket;
        this.socketChannel = socket.getChannel();
        if (this.socketChannel != null) {
            try {
                this.socketChannel.configureBlocking(!b);
            }
            catch (IOException cause) {
                RuntimeException ex = new RuntimeException();
                ex.initCause(cause);
                throw ex;
            }
        }
        this.acceptor = acceptor;
        this.serverRequestMap = Collections.synchronizedMap(new HashMap());
        this.isServer = true;
        this.state = 2;
    }
    
    public SocketOrChannelConnectionImpl(ORB orb, Acceptor acceptor, Socket socket) {
        this(orb, acceptor, socket, socket.getChannel() != null && orb.getORBData().connectionSocketUseSelectThreadToWait(), socket.getChannel() != null && orb.getORBData().connectionSocketUseWorkerThreadForEvent());
    }
    
    public boolean shouldRegisterReadEvent() {
        return true;
    }
    
    public boolean shouldRegisterServerReadEvent() {
        return true;
    }
    
    public boolean read() {
        try {
            if (this.orb.transportDebugFlag) {
                this.dprint(".read->: " + this);
            }
            CorbaMessageMediator bits = this.readBits();
            if (bits != null) {
                return this.dispatch(bits);
            }
            return true;
        }
        finally {
            if (this.orb.transportDebugFlag) {
                this.dprint(".read<-: " + this);
            }
        }
    }
    
    protected CorbaMessageMediator readBits() {
        try {
            if (this.orb.transportDebugFlag) {
                this.dprint(".readBits->: " + this);
            }
            MessageMediator messageMediator;
            if (this.contactInfo != null) {
                messageMediator = this.contactInfo.createMessageMediator(this.orb, this);
            }
            else if (this.acceptor != null) {
                messageMediator = this.acceptor.createMessageMediator(this.orb, this);
            }
            else {
                throw new RuntimeException("SocketOrChannelConnectionImpl.readBits");
            }
            CorbaMessageMediator corbaMessageMediator = (CorbaMessageMediator)messageMediator;
        }
        catch (ThreadDeath threadDeath) {
            if (this.orb.transportDebugFlag) {
                this.dprint(".readBits: " + this + ": ThreadDeath: " + threadDeath, threadDeath);
            }
            try {
                this.purgeCalls(this.wrapper.connectionAbort(threadDeath), false, false);
            }
            catch (Throwable obj) {
                if (this.orb.transportDebugFlag) {
                    this.dprint(".readBits: " + this + ": purgeCalls: Throwable: " + obj, obj);
                }
            }
            throw threadDeath;
        }
        catch (Throwable threadDeath) {
            if (this.orb.transportDebugFlag) {
                this.dprint(".readBits: " + this + ": Throwable: " + threadDeath, threadDeath);
            }
            try {
                if ((threadDeath instanceof INTERNAL)) {
                    this.sendMessageError(GIOPVersion.DEFAULT_VERSION);
                }
            }
            catch (IOException obj2) {
                if (this.orb.transportDebugFlag) {
                    this.dprint(".readBits: " + this + ": sendMessageError: IOException: " + obj2, obj2);
                }
            }
            Selector selector = this.orb.getTransportManager().getSelector(0);
            if (selector != null) {
                selector.unregisterForEvent(this);
            }
            this.purgeCalls(this.wrapper.connectionAbort(threadDeath), true, false);
        }
        finally {
            if (this.orb.transportDebugFlag) {
                this.dprint(".readBits<-: " + this);
            }
        }
        return null;
    }
    
    protected CorbaMessageMediator finishReadingBits(MessageMediator messageMediator) {
        try {
            if (this.orb.transportDebugFlag) {
                this.dprint(".finishReadingBits->: " + this);
            }
            if (this.contactInfo != null) {
                messageMediator = this.contactInfo.finishCreatingMessageMediator(this.orb, this, messageMediator);
            }
            else if (this.acceptor != null) {
                messageMediator = this.acceptor.finishCreatingMessageMediator(this.orb, this, messageMediator);
            }
            else {
                throw new RuntimeException("SocketOrChannelConnectionImpl.finishReadingBits");
            }
            CorbaMessageMediator corbaMessageMediator = (CorbaMessageMediator)messageMediator;
        }
        catch (ThreadDeath threadDeath) {
            if (this.orb.transportDebugFlag) {
                this.dprint(".finishReadingBits: " + this + ": ThreadDeath: " + threadDeath, threadDeath);
            }
            try {
                this.purgeCalls(this.wrapper.connectionAbort(threadDeath), false, false);
            }
            catch (Throwable obj) {
                if (this.orb.transportDebugFlag) {
                    this.dprint(".finishReadingBits: " + this + ": purgeCalls: Throwable: " + obj, obj);
                }
            }
            throw threadDeath;
        }
        catch (Throwable threadDeath) {
            if (this.orb.transportDebugFlag) {
                this.dprint(".finishReadingBits: " + this + ": Throwable: " + threadDeath, threadDeath);
            }
            try {
                if ((threadDeath instanceof INTERNAL)) {
                    this.sendMessageError(GIOPVersion.DEFAULT_VERSION);
                }
            }
            catch (IOException obj2) {
                if (this.orb.transportDebugFlag) {
                    this.dprint(".finishReadingBits: " + this + ": sendMessageError: IOException: " + obj2, obj2);
                }
            }
            this.orb.getTransportManager().getSelector(0).unregisterForEvent(this);
            this.purgeCalls(this.wrapper.connectionAbort(threadDeath), true, false);
        }
        finally {
            if (this.orb.transportDebugFlag) {
                this.dprint(".finishReadingBits<-: " + this);
            }
        }
        return null;
    }
    
    protected boolean dispatch(CorbaMessageMediator corbaMessageMediator) {
        try {
            if (this.orb.transportDebugFlag) {
                this.dprint(".dispatch->: " + this);
            }
            return handleRequest = corbaMessageMediator.getProtocolHandler().handleRequest(corbaMessageMediator);
        }
        catch (ThreadDeath threadDeath) {
            if (this.orb.transportDebugFlag) {
                this.dprint(".dispatch: ThreadDeath", threadDeath);
            }
            try {
                this.purgeCalls(this.wrapper.connectionAbort(threadDeath), false, false);
            }
            catch (Throwable t) {
                if (this.orb.transportDebugFlag) {
                    this.dprint(".dispatch: purgeCalls: Throwable", t);
                }
            }
            throw threadDeath;
        }
        catch (Throwable threadDeath) {
            if (this.orb.transportDebugFlag) {
                this.dprint(".dispatch: Throwable", threadDeath);
            }
            try {
                if ((threadDeath instanceof INTERNAL)) {
                    this.sendMessageError(GIOPVersion.DEFAULT_VERSION);
                }
            }
            catch (IOException ex) {
                if (this.orb.transportDebugFlag) {
                    this.dprint(".dispatch: sendMessageError: IOException", ex);
                }
            }
            this.purgeCalls(this.wrapper.connectionAbort(threadDeath), false, false);
        }
        finally {
            if (this.orb.transportDebugFlag) {
                this.dprint(".dispatch<-: " + this);
            }
        }
        return true;
    }
    
    public boolean shouldUseDirectByteBuffers() {
        return this.getSocketChannel() != null;
    }
    
    public ByteBuffer read(int n, int n2, int n3, long n4) throws IOException {
        if (this.shouldUseDirectByteBuffers()) {
            ByteBuffer byteBuffer = this.orb.getByteBufferPool().getByteBuffer(n);
            if (this.orb.transportDebugFlag) {
                int identityHashCode = System.identityHashCode(byteBuffer);
                StringBuffer sb = new StringBuffer(80);
                sb.append(".read: got ByteBuffer id (");
                sb.append(identityHashCode).append(") from ByteBufferPool.");
                this.dprint(sb.toString());
            }
            byteBuffer.position(n2);
            byteBuffer.limit(n);
            this.readFully(byteBuffer, n3, n4);
            return byteBuffer;
        }
        else {
            byte[] array = new byte[n];
            this.readFully(this.getSocket().getInputStream(), array, n2, n3, n4);
            ByteBuffer wrap = ByteBuffer.wrap(array);
            wrap.limit(n);
            return wrap;
        }
    }
    
    public ByteBuffer read(ByteBuffer byteBuffer, int n, int n2, long n3) throws IOException {
        int n4 = n + n2;
        if (this.shouldUseDirectByteBuffers()) {
            if (!byteBuffer.isDirect()) {
                throw this.wrapper.unexpectedNonDirectByteBufferWithChannelSocket();
            }
            if (n4 > byteBuffer.capacity()) {
                if (this.orb.transportDebugFlag) {
                    int identityHashCode = System.identityHashCode(byteBuffer);
                    StringBuffer sb = new StringBuffer(80);
                    sb.append(".read: releasing ByteBuffer id (").append(identityHashCode).append(") to ByteBufferPool.");
                    this.dprint(sb.toString());
                }
                this.orb.getByteBufferPool().releaseByteBuffer(byteBuffer);
                byteBuffer = this.orb.getByteBufferPool().getByteBuffer(n4);
            }
            byteBuffer.position(n);
            byteBuffer.limit(n4);
            this.readFully(byteBuffer, n2, n3);
            byteBuffer.position(0);
            byteBuffer.limit(n4);
            return byteBuffer;
        }
        else {
            if (byteBuffer.isDirect()) {
                throw this.wrapper.unexpectedDirectByteBufferWithNonChannelSocket();
            }
            byte[] array = new byte[n4];
            this.readFully(this.getSocket().getInputStream(), array, n, n2, n3);
            return ByteBuffer.wrap(array);
        }
    }
    
    public void readFully(ByteBuffer byteBuffer, int value, long value2) throws IOException {
        int value3 = 0;
        long millis = (long)this.readTimeouts.get_initial_time_to_wait();
        long value4 = 0L;
        do {
            int read = this.getSocketChannel().read(byteBuffer);
            if (read < 0) {
                throw new IOException("End-of-stream");
            }
            if (read == 0) {
                try {
                    Thread.sleep(millis);
                    value4 += millis;
                    millis = (long)((double)millis * this.readTimeouts.get_backoff_factor());
                }
                catch (InterruptedException ex) {
                    if (this.orb.transportDebugFlag) {
                        this.dprint("readFully(): unexpected exception " + ex.toString());
                    }
                    else {
                        continue;
                    }
                }
            }
            else {
                value3 += read;
            }
        } while (value3 < value && value4 < value2);
        if (value3 < value && value4 >= value2) {
            throw this.wrapper.transportReadTimeoutExceeded(new Integer(value), new Integer(value3), new Long(value2), new Long(value4));
        }
        this.getConnectionCache().stampTime(this);
        return;
    }
    
    public void readFully(InputStream inputStream, byte[] b, int n, int value, long value2) throws IOException {
        int value3 = 0;
        long millis = (long)this.readTimeouts.get_initial_time_to_wait();
        long value4 = 0L;
        do {
            int read = inputStream.read(b, n + value3, value - value3);
            if (read < 0) {
                throw new IOException("End-of-stream");
            }
            if (read == 0) {
                try {
                    Thread.sleep(millis);
                    value4 += millis;
                    millis = (long)((double)millis * this.readTimeouts.get_backoff_factor());
                }
                catch (InterruptedException ex) {
                    if (this.orb.transportDebugFlag) {
                        this.dprint("readFully(): unexpected exception " + ex.toString());
                    }
                    else {
                        continue;
                    }
                }
            }
            else {
                value3 += read;
            }
        } while (value3 < value && value4 < value2);
        if (value3 < value && value4 >= value2) {
            throw this.wrapper.transportReadTimeoutExceeded(new Integer(value), new Integer(value3), new Long(value2), new Long(value4));
        }
        this.getConnectionCache().stampTime(this);
        return;
    }
    
    public void write(ByteBuffer byteBuffer) throws IOException {
        if (this.shouldUseDirectByteBuffers()) {
            do {
                this.getSocketChannel().write(byteBuffer);
            } while (byteBuffer.hasRemaining());
        }
        else {
            if (!byteBuffer.hasArray()) {
                throw this.wrapper.unexpectedDirectByteBufferWithNonChannelSocket();
            }
            this.getSocket().getOutputStream().write(byteBuffer.array(), 0, byteBuffer.limit());
            this.getSocket().getOutputStream().flush();
        }
        this.getConnectionCache().stampTime(this);
    }
    
    public synchronized void close() {
        try {
            if (this.orb.transportDebugFlag) {
                this.dprint(".close->: " + this);
            }
            this.writeLock();
            if (this.isBusy()) {
                this.writeUnlock();
                if (this.orb.transportDebugFlag) {
                    this.dprint(".close: isBusy so no close: " + this);
                }
                return;
            }
            else {
                try {
                    try {
                        this.sendCloseConnection(GIOPVersion.V1_0);
                    }
                    catch (Throwable t) {
                        this.wrapper.exceptionWhenSendingCloseConnection(t);
                    }
                    synchronized (this.stateEvent) {
                        this.state = 3;
                        this.stateEvent.notifyAll();
                    }
                    this.purgeCalls(this.wrapper.connectionRebind(), false, true);
                }
                catch (Exception ex) {
                    if (this.orb.transportDebugFlag) {
                        this.dprint(".close: exception: " + this, ex);
                    }
                }
                try {
                    Selector selector = this.orb.getTransportManager().getSelector(0);
                    if (selector != null) {
                        selector.unregisterForEvent(this);
                    }
                    if (this.socketChannel != null) {
                        this.socketChannel.close();
                    }
                    this.socket.close();
                }
                catch (IOException ex2) {
                    if (this.orb.transportDebugFlag) {
                        this.dprint(".close: " + this, ex2);
                    }
                }
                this.closeConnectionResources();
            }
        }
        finally {
            if (this.orb.transportDebugFlag) {
                this.dprint(".close<-: " + this);
            }
        }
    }
    
    public void closeConnectionResources() {
        if (this.orb.transportDebugFlag) {
            this.dprint(".closeConnectionResources->: " + this);
        }
        Selector selector = this.orb.getTransportManager().getSelector(0);
        if (selector != null) {
            selector.unregisterForEvent(this);
        }
        try {
            if (this.socketChannel != null) {
                this.socketChannel.close();
            }
            if (this.socket != null && !this.socket.isClosed()) {
                this.socket.close();
            }
        }
        catch (IOException ex) {
            if (this.orb.transportDebugFlag) {
                this.dprint(".closeConnectionResources: " + this, ex);
            }
        }
        if (this.orb.transportDebugFlag) {
            this.dprint(".closeConnectionResources<-: " + this);
        }
    }
    
    public Acceptor getAcceptor() {
        return this.acceptor;
    }
    
    public ContactInfo getContactInfo() {
        return this.contactInfo;
    }
    
    public EventHandler getEventHandler() {
        return this;
    }
    
    public OutputObject createOutputObject(MessageMediator messageMediator) {
        throw new RuntimeException("*****SocketOrChannelConnectionImpl.createOutputObject - should not be called.");
    }
    
    public boolean isServer() {
        return this.isServer;
    }
    
    public boolean isBusy() {
        return this.serverRequestCount > 0 || this.getResponseWaitingRoom().numberRegistered() > 0;
    }
    
    public long getTimeStamp() {
        return this.timeStamp;
    }
    
    public void setTimeStamp(long timeStamp) {
        this.timeStamp = timeStamp;
    }
    
    public void setState(String s) {
        synchronized (this.stateEvent) {
            if (s.equals("ESTABLISHED")) {
                this.state = 2;
                this.stateEvent.notifyAll();
            }
        }
    }
    
    public void writeLock() {
        try {
            if (SocketOrChannelConnectionImpl.dprintWriteLocks && this.orb.transportDebugFlag) {
                this.dprint(".writeLock->: " + this);
            }
        Label_0370:
            while (true) {
                switch (this.state) {
                    case 1: {
                        synchronized (this.stateEvent) {
                            if (this.state == 1) {
                                try {
                                    this.stateEvent.wait();
                                }
                                catch (InterruptedException ex) {
                                    if (this.orb.transportDebugFlag) {
                                        this.dprint(".writeLock: OPENING InterruptedException: " + this);
                                    }
                                }
                            }
                        }
                        continue;
                    }
                    case 2: {
                        synchronized (this.writeEvent) {
                            if (!this.writeLocked) {
                                this.writeLocked = true;
                                (this.writeEvent);
                                return;
                            }
                            try {
                                while (this.state == 2 && this.writeLocked) {
                                    this.writeEvent.wait(100L);
                                }
                            }
                            catch (InterruptedException ex2) {
                                if (this.orb.transportDebugFlag) {
                                    this.dprint(".writeLock: ESTABLISHED InterruptedException: " + this);
                                }
                            }
                        }
                        continue;
                    }
                    case 5: {
                        synchronized (this.stateEvent) {
                            if (this.state != 5) {
                                continue;
                            }
                            throw this.wrapper.writeErrorSend();
                        }
                    }
                    case 4: {
                        synchronized (this.stateEvent) {
                            if (this.state != 4) {
                                continue;
                            }
                            throw this.wrapper.connectionCloseRebind();
                        }
                        break Label_0370;
                    }
                    default: {
                        break Label_0370;
                    }
                }
            }
            if (this.orb.transportDebugFlag) {
                this.dprint(".writeLock: default: " + this);
            }
            throw new RuntimeException(".writeLock: bad state");
        }
        finally {
            if (SocketOrChannelConnectionImpl.dprintWriteLocks && this.orb.transportDebugFlag) {
                this.dprint(".writeLock<-: " + this);
            }
        }
    }
    
    public void writeUnlock() {
        try {
            if (SocketOrChannelConnectionImpl.dprintWriteLocks && this.orb.transportDebugFlag) {
                this.dprint(".writeUnlock->: " + this);
            }
            synchronized (this.writeEvent) {
                this.writeLocked = false;
                this.writeEvent.notify();
            }
        }
        finally {
            if (SocketOrChannelConnectionImpl.dprintWriteLocks && this.orb.transportDebugFlag) {
                this.dprint(".writeUnlock<-: " + this);
            }
        }
    }
    
    public void sendWithoutLock(OutputObject outputObject) {
        try {
            ((CDROutputObject)outputObject).writeTo(this);
        }
        catch (IOException ex) {
            COMM_FAILURE writeErrorSend = this.wrapper.writeErrorSend(ex);
            this.purgeCalls(writeErrorSend, false, true);
            throw writeErrorSend;
        }
    }
    
    public void registerWaiter(MessageMediator messageMediator) {
        this.responseWaitingRoom.registerWaiter(messageMediator);
    }
    
    public void unregisterWaiter(MessageMediator messageMediator) {
        this.responseWaitingRoom.unregisterWaiter(messageMediator);
    }
    
    public InputObject waitForResponse(MessageMediator messageMediator) {
        return this.responseWaitingRoom.waitForResponse(messageMediator);
    }
    
    public void setConnectionCache(ConnectionCache connectionCache) {
        this.connectionCache = connectionCache;
    }
    
    public ConnectionCache getConnectionCache() {
        return this.connectionCache;
    }
    
    @Override
    public void setUseSelectThreadToWait(boolean useSelectThreadToWait) {
        this.useSelectThreadToWait = useSelectThreadToWait;
        this.setReadGiopHeaderOnly(this.shouldUseSelectThreadToWait());
    }
    
    @Override
    public void handleEvent() {
        if (this.orb.transportDebugFlag) {
            this.dprint(".handleEvent->: " + this);
        }
        this.getSelectionKey().interestOps(this.getSelectionKey().interestOps() & ~this.getInterestOps());
        if (this.shouldUseWorkerThreadForEvent()) {
            Object obj = null;
            try {
                int threadPoolToUse = 0;
                if (this.shouldReadGiopHeaderOnly()) {
                    this.partialMessageMediator = this.readBits();
                    threadPoolToUse = this.partialMessageMediator.getThreadPoolToUse();
                }
                if (this.orb.transportDebugFlag) {
                    this.dprint(".handleEvent: addWork to pool: " + threadPoolToUse);
                }
                this.orb.getThreadPoolManager().getThreadPool(threadPoolToUse).getWorkQueue(0).addWork(this.getWork());
            }
            catch (NoSuchThreadPoolException ex) {
                obj = ex;
            }
            catch (NoSuchWorkQueueException ex2) {
                obj = ex2;
            }
            if (obj != null) {
                if (this.orb.transportDebugFlag) {
                    this.dprint(".handleEvent: " + obj);
                }
                INTERNAL o = new INTERNAL("NoSuchThreadPoolException");
                o.initCause(obj);
                throw o;
            }
        }
        else {
            if (this.orb.transportDebugFlag) {
                this.dprint(".handleEvent: doWork");
            }
            this.getWork().doWork();
        }
        if (this.orb.transportDebugFlag) {
            this.dprint(".handleEvent<-: " + this);
        }
    }
    
    public SelectableChannel getChannel() {
        return this.socketChannel;
    }
    
    public int getInterestOps() {
        return 1;
    }
    
    public Connection getConnection() {
        return this;
    }
    
    public String getName() {
        return this.toString();
    }
    
    public void doWork() {
        try {
            if (this.orb.transportDebugFlag) {
                this.dprint(".doWork->: " + this);
            }
            if (!this.shouldReadGiopHeaderOnly()) {
                this.read();
            }
            else {
                CorbaMessageMediator finishReadingBits = this.finishReadingBits(this.getPartialMessageMediator());
                if (finishReadingBits != null) {
                    this.dispatch(finishReadingBits);
                }
            }
        }
        catch (Throwable obj) {
            if (this.orb.transportDebugFlag) {
                this.dprint(".doWork: ignoring Throwable: " + obj + " " + this);
            }
        }
        finally {
            if (this.orb.transportDebugFlag) {
                this.dprint(".doWork<-: " + this);
            }
        }
    }
    
    public void setEnqueueTime(long enqueueTime) {
        this.enqueueTime = enqueueTime;
    }
    
    public long getEnqueueTime() {
        return this.enqueueTime;
    }
    
    public boolean shouldReadGiopHeaderOnly() {
        return this.shouldReadGiopHeaderOnly;
    }
    
    protected void setReadGiopHeaderOnly(boolean shouldReadGiopHeaderOnly) {
        this.shouldReadGiopHeaderOnly = shouldReadGiopHeaderOnly;
    }
    
    public ResponseWaitingRoom getResponseWaitingRoom() {
        return this.responseWaitingRoom;
    }
    
    public void serverRequestMapPut(int value, CorbaMessageMediator corbaMessageMediator) {
        this.serverRequestMap.put(new Integer(value), corbaMessageMediator);
    }
    
    public CorbaMessageMediator serverRequestMapGet(int value) {
        return (CorbaMessageMediator)this.serverRequestMap.get(new Integer(value));
    }
    
    public void serverRequestMapRemove(int value) {
        this.serverRequestMap.remove(new Integer(value));
    }
    
    public Socket getSocket() {
        return this.socket;
    }
    
    public synchronized void serverRequestProcessingBegins() {
        ++this.serverRequestCount;
    }
    
    public synchronized void serverRequestProcessingEnds() {
        --this.serverRequestCount;
    }
    
    public synchronized int getNextRequestId() {
        return this.requestId++;
    }
    
    public ORB getBroker() {
        return this.orb;
    }
    
    public CodeSetComponentInfo.CodeSetContext getCodeSetContext() {
        if (this.codeSetContext == null) {
            synchronized (this) {
                return this.codeSetContext;
            }
        }
        return this.codeSetContext;
    }
    
    public synchronized void setCodeSetContext(CodeSetComponentInfo.CodeSetContext codeSetContext) {
        if (this.codeSetContext == null) {
            if (OSFCodeSetRegistry.lookupEntry(codeSetContext.getCharCodeSet()) == null || OSFCodeSetRegistry.lookupEntry(codeSetContext.getWCharCodeSet()) == null) {
                throw this.wrapper.badCodesetsFromClient();
            }
            this.codeSetContext = codeSetContext;
        }
    }
    
    public MessageMediator clientRequestMapGet(int n) {
        return this.responseWaitingRoom.getMessageMediator(n);
    }
    
    public void clientReply_1_1_Put(MessageMediator clientReply_1_1) {
        this.clientReply_1_1 = clientReply_1_1;
    }
    
    public MessageMediator clientReply_1_1_Get() {
        return this.clientReply_1_1;
    }
    
    public void clientReply_1_1_Remove() {
        this.clientReply_1_1 = null;
    }
    
    public void serverRequest_1_1_Put(MessageMediator serverRequest_1_1) {
        this.serverRequest_1_1 = serverRequest_1_1;
    }
    
    public MessageMediator serverRequest_1_1_Get() {
        return this.serverRequest_1_1;
    }
    
    public void serverRequest_1_1_Remove() {
        this.serverRequest_1_1 = null;
    }
    
    protected String getStateString(int n) {
        synchronized (this.stateEvent) {
            switch (n) {
                case 1: {
                    return "OPENING";
                }
                case 2: {
                    return "ESTABLISHED";
                }
                case 3: {
                    return "CLOSE_SENT";
                }
                case 4: {
                    return "CLOSE_RECVD";
                }
                case 5: {
                    return "ABORT";
                }
                default: {
                    return "???";
                }
            }
        }
    }
    
    public synchronized boolean isPostInitialContexts() {
        return this.postInitialContexts;
    }
    
    public synchronized void setPostInitialContexts() {
        this.postInitialContexts = true;
    }
    
    public void purgeCalls(SystemException ex, boolean b, boolean b2) {
        int minor = ex.minor;
        try {
            if (this.orb.transportDebugFlag) {
                this.dprint(".purgeCalls->: " + minor + "/" + b + "/" + b2 + " " + this);
            }
            synchronized (this.stateEvent) {
                if (this.state == 5 || this.state == 4) {
                    if (this.orb.transportDebugFlag) {
                        this.dprint(".purgeCalls: exiting since state is: " + this.getStateString(this.state) + " " + this);
                    }
                    return;
                }
            }
            try {
                if (!b2) {
                    this.writeLock();
                }
            }
            catch (SystemException obj) {
                if (this.orb.transportDebugFlag) {
                    this.dprint(".purgeCalls: SystemException" + obj + "; continuing " + this);
                }
            }
            synchronized (this.stateEvent) {
                if (minor == 1398079697) {
                    this.state = 4;
                    ex.completed = CompletionStatus.COMPLETED_NO;
                }
                else {
                    this.state = 5;
                    ex.completed = CompletionStatus.COMPLETED_MAYBE;
                }
                this.stateEvent.notifyAll();
            }
            try {
                this.socket.getInputStream().close();
                this.socket.getOutputStream().close();
                this.socket.close();
            }
            catch (Exception obj2) {
                if (this.orb.transportDebugFlag) {
                    this.dprint(".purgeCalls: Exception closing socket: " + obj2 + " " + this);
                }
            }
            this.responseWaitingRoom.signalExceptionToAllWaiters(ex);
        }
        finally {
            if (this.contactInfo != null) {
                ((OutboundConnectionCache)this.getConnectionCache()).remove(this.contactInfo);
            }
            else if (this.acceptor != null) {
                ((InboundConnectionCache)this.getConnectionCache()).remove(this);
            }
            this.writeUnlock();
            if (this.orb.transportDebugFlag) {
                this.dprint(".purgeCalls<-: " + minor + "/" + b + "/" + b2 + " " + this);
            }
        }
    }
    
    public void sendCloseConnection(GIOPVersion giopVersion) throws IOException {
        this.sendHelper(giopVersion, MessageBase.createCloseConnection(giopVersion));
    }
    
    public void sendMessageError(GIOPVersion giopVersion) throws IOException {
        this.sendHelper(giopVersion, MessageBase.createMessageError(giopVersion));
    }
    
    public void sendCancelRequest(GIOPVersion giopVersion, int n) throws IOException {
        this.sendHelper(giopVersion, MessageBase.createCancelRequest(giopVersion, n));
    }
    
    protected void sendHelper(GIOPVersion giopVersion, Message message) throws IOException {
        CDROutputObject cdrOutputObject = OutputStreamFactory.newCDROutputObject(this.orb, null, giopVersion, this, message, (byte)1);
        message.write(cdrOutputObject);
        cdrOutputObject.writeTo(this);
    }
    
    public void sendCancelRequestWithLock(GIOPVersion giopVersion, int n) throws IOException {
        this.writeLock();
        try {
            this.sendCancelRequest(giopVersion, n);
        }
        finally {
            this.writeUnlock();
        }
    }
    
    public final void setCodeBaseIOR(IOR codeBaseServerIOR) {
        this.codeBaseServerIOR = codeBaseServerIOR;
    }
    
    public final IOR getCodeBaseIOR() {
        return this.codeBaseServerIOR;
    }
    
    public final CodeBase getCodeBase() {
        return this.cachedCodeBase;
    }
    
    protected void setReadTimeouts(ReadTimeouts readTimeouts) {
        this.readTimeouts = readTimeouts;
    }
    
    protected void setPartialMessageMediator(CorbaMessageMediator partialMessageMediator) {
        this.partialMessageMediator = partialMessageMediator;
    }
    
    protected CorbaMessageMediator getPartialMessageMediator() {
        return this.partialMessageMediator;
    }
    
    public String toString() {
        synchronized (this.stateEvent) {
            return "SocketOrChannelConnectionImpl[ " + ((this.socketChannel == null) ? this.socket.toString() : this.socketChannel.toString()) + " " + this.getStateString(this.state) + " " + this.shouldUseSelectThreadToWait() + " " + this.shouldUseWorkerThreadForEvent() + " " + this.shouldReadGiopHeaderOnly() + "]";
        }
    }
    
    public void dprint(String s) {
        ORBUtility.dprint("SocketOrChannelConnectionImpl", s);
    }
    
    protected void dprint(String s, Throwable t) {
        this.dprint(s);
        t.printStackTrace(System.out);
    }
    
    static {
        SocketOrChannelConnectionImpl.dprintWriteLocks = false;
    }
}
